from djitellopy import Tello
import numpy as np
import cv2
import time
import threading

def detecting_lines(frame):
    try:
        contrast_frame = cv2.convertScaleAbs(frame, alpha=2, beta=0)
        gray_frame = cv2.cvtColor(contrast_frame, cv2.COLOR_BGR2GRAY)
        # gray_frame = cv2.cvtColor(frame, cv2.COLOR_BGR2GRAY)

        _, binary_frame = cv2.threshold(gray_frame, 150, 255, cv2.THRESH_BINARY) # 127 might be 100.

        edges = cv2.Canny(binary_frame, 100, 200)

        kernel = np.ones((8, 8), np.uint8)
        dilated_edges = cv2.dilate(edges, kernel, iterations = 1)

        lines = cv2.HoughLinesP(dilated_edges, 1, np.pi / 180, threshold = 100, minLineLength = 100, maxLineGap = 8)

        total_line_detected = 0
        
        if lines is not None:
            for index, line in enumerate(lines):
                x1, y1, x2, y2 = line[0]
                
                line_length = np.sqrt((x2 - x1)**2 + (y2 - y1)**2)
                
                if line_length > 75:
                    # print(f"Detected line {index} coordinates: ({x1}, {y1}, {x2}, {y2})")
                    cv2.line(frame, (x1, y1), (x2, y2), (0, 255, 0), 3)
                    total_line_detected += 1
        
        return frame, binary_frame, total_line_detected
    
    except Exception as e:
        print(f"Error in detecting lines: {e}")

def tello_visualization_with_photo(tello_drone):
    try:
        while True:
            frame_read = tello_drone.get_frame_read()
            frame = frame_read.frame

            processed_frame, edge_frame, detected_line_count = detecting_lines(frame = frame)
            
            cv2.imshow("Original frame with detected lines", processed_frame)
            if edge_frame is not None:
                cv2.imshow("Edges", edge_frame)

            key = cv2.waitKey(3) & 0xFF

            if key == ord('c'):
                timestamp = time.strftime("%Y%m%d-%H%M%S")
                
                filenameFirst = f"Drone_{timestamp}_BINARY_FRAME.jpg"
                cv2.imwrite(filenameFirst, edge_frame)
                print(f"Photo captured and saved as: {filenameFirst}.")
                
                filenameSecond = f"Drone_{timestamp}_PROCESSED_FRAME.jpg"
                cv2.imwrite(filenameSecond, processed_frame)
                print(f"Photo captured and saved as: {filenameSecond}.")

            if key == ord('s'):
                tello_drone.streamoff()
                tello_drone.land()
                tello_drone.end()
                break

            if key == ord('q'):
                break

            if key == ord('e'):
                print("Emergency stop triggered. Landing drone...")
                tello_drone.land()
                break

            battery_indicator = 10

            if tello_drone.get_battery() < battery_indicator:
                print(f"Battery lower than {battery_indicator}! Landing immediately...")
                tello_drone.land()
                tello_drone.end()
                break

            time.sleep(0.033)

    except Exception as e:
        print(f"Error in video streaming: {e}")

    finally:
        print("TOTAL DETECTED LINES: ", detected_line_count)
        cv2.destroyAllWindows()

def drone_movements(tello_drone):
    try:
        print("Taking off...")
        tello_drone.takeoff()

        time.sleep(3)

        print("Moving up...")
        tello_drone.move_up(112)

        time.sleep(7)
        '''
        print("Rotating clockwise...")
        tello_drone.rotate_clockwise(45)

        time.sleep(3)

        print("Rotating counter clockwise...")
        tello_drone.rotate_counter_clockwise(45)

        time.sleep(3)
        '''
        print("Landing...")
        tello_drone.land()
        
    except Exception as e:
        print(f"Error in drone movements: {e}")

    finally:
        print("Drone movements completed.")

def main():
    try:
        drone = Tello()

        drone.connect()
        print(f"Battery: {drone.get_battery()}%")

        drone.streamon()

        video_thread = threading.Thread(target = tello_visualization_with_photo, args = (drone, ))
        movement_thread = threading.Thread(target = drone_movements, args = (drone, ))

        video_thread.start()
        movement_thread.start()

        video_thread.join()
        movement_thread.join()

    except Exception as e:
        print(f"Unexpected error (MAIN function): {e}")

    except ConnectionError as e:
        print(f"Connection error: {e}")

    finally:
        print("CLeaning up...")
        drone.streamoff()
        drone.end()
    
if __name__ == "__main__":
    main()